from loopy.target import (TargetBase,
                          ASTBuilderBase,
                          DummyHostASTBuilder,
                          )
from loopy.target.c import CASTBuilder
from loopy.target.c.codegen.expression import ExpressionToCMapper

_registry = {'float32': 'float',
             'int32': 'int',
             'float64': 'double',
             'string': 'std::string',
             'str': 'std::string'}


class MyMapper(ExpressionToCMapper):
    def map_subscript(self, expr, enclosing_prec, type_context):
        from dune.perftool.generation.loopy import UnmanagedTemporaryVariable
        if isinstance(self.find_array(expr), UnmanagedTemporaryVariable):
            ret = str(expr.aggregate)
            from pymbolic.primitives import Variable
            if isinstance(expr.index, Variable):
                ret = ret + '[{}]'.format(str(expr.index))
            else:
                for i in expr.index:
                    ret = ret + '[{}]'.format(str(i))
            return ret
        else:
            return ExpressionToCMapper.map_subscript(self, expr, enclosing_prec, type_context)


class DuneASTBuilder(CASTBuilder):
    def get_expression_to_code_mapper(self, codegen_state):
        return MyMapper(codegen_state)

    def get_temporary_decls(self, codegen_state, schedule_index):
        # Currently we do *not* want to build the temporary declarations
        # through loopy, as this would involve lots of work on a proper
        # type system!
        return []

    def emit_sequential_loop(self, codegen_state, iname, iname_dtype, static_lbound, static_ubound, inner):
        # Some of our loops are special as they use PDELab specific concepts.
        # Fortunately those loops are tied to specific inames.
        from dune.perftool.pdelab.quadrature import quadrature_iname
        if iname == quadrature_iname():
            from cgen import CustomLoop
            from dune.perftool.pdelab.quadrature import quadrature_loop_statement
            loop_stmt = quadrature_loop_statement()
            return CustomLoop("for({})".format(loop_stmt), inner)
        else:
            return CASTBuilder.emit_sequential_loop(self, codegen_state, iname, iname_dtype, static_lbound, static_ubound, inner)


class DuneTarget(TargetBase):
    def __init__(self):
        # Set fortran_abi to allow reusing CASTBuilder for the moment
        self.fortran_abi = False

    def split_kernel_at_global_barriers(self):
        return False

    def get_host_ast_builder(self):
        return DummyHostASTBuilder(self)

    def get_device_ast_builder(self):
        return DuneASTBuilder(self)

    def dtype_to_typename(self, dtype):
        # For now, we do this the simplest possible way
        return _registry[dtype.dtype.name]

    def is_vector_dtype(self, dtype):
        return False
