from dune.perftool.generation import (class_member,
                                      constructor_block,
                                      domain,
                                      iname,
                                      instruction,
                                      no_caching,
                                      preamble,
                                      )
from dune.perftool.pdelab.localoperator import lop_template_range_field


class AMatrix(object):
    def __init__(self, a_matrix, m, n):
        self.a_matrix = a_matrix
        self.m = m
        self.n = n

    def __str__(self):
        return "a_matrix: {}\nm: {}\nn: {}".format(self.a_matrix, self.m, self.n)


class _GlobalBufferCounter(object):
    counter = 0

    def get(self):
        _GlobalBufferCounter.counter = _GlobalBufferCounter.counter + 1
        return _GlobalBufferCounter.counter


@constructor_block("operator")
def construct_buffer(name, a_matrices):
    # TODO Free memory in destructor!
    range_field = lop_template_range_field()

    # Upper bound for the number of elements in the buffer. Note: This
    # is usually much too large and could be improved.
    max_size = 1
    for mat in a_matrices:
        max_size *= max(mat.n, mat.m)

    from dune.perftool.sumfact.amatrix import name_alignment
    alignment = name_alignment()
    return ["{} = ({}*) aligned_malloc({}*sizeof({}), {});".format(name, range_field, max_size, range_field, alignment),
            "if ({}==0) throw std::bad_alloc();".format(name)]


def type_buffer():
    range_field = lop_template_range_field()
    return "mutable {}*".format(range_field)


@class_member("operator")
def define_buffer(name, a_matrices):
    buffer_type = type_buffer()
    construct_buffer(name, a_matrices)
    return "{} {};".format(buffer_type, name)


def name_buffer(a_matrices):
    name = "sumfact_buffer_" + str(_GlobalBufferCounter().get())
    define_buffer(name, a_matrices)
    return name


# TODO Avoid caching?
class _GlobalSumfactCounter(object):
    counter = 0

    def get(self):
        _GlobalSumfactCounter.counter = _GlobalSumfactCounter.counter + 1
        return _GlobalSumfactCounter.counter


@iname
def _sumfact_rows_iname(a_matrix, count):
    # TODO: Avoid caching?
    # count = _GlobalSumfactCounter().get()
    name = "sf_cols_" + str(count)
    domain(name, a_matrix.m)
    return name


def sumfact_rows_iname(a_matrix):
    count = _GlobalSumfactCounter().get()
    name = _sumfact_rows_iname(a_matrix, count)
    return name


@iname
def _sumfact_collumns_iname(input_collumns, count):
    # TODO: Avoid caching?
    # count = _GlobalSumfactCounter().get()
    name = "sf_rows_" + str(count)
    domain(name, int(input_collumns))
    return name


def sumfact_collumns_iname(input_collumns):
    count = _GlobalSumfactCounter().get()
    name = _sumfact_collumns_iname(input_collumns, count)
    return name


def _sumfact_kernel(a_matrix, inp, out, input_collumns,
                    first, element, name, container, restriction, component):
    """Sumfactorization kernel computing out = a_matrix * inp

    Rotation step is directly incuded. This can be done by treating
    inp as a matrix in column major storage and out as a row major
    matrix. With Theta of size m_lxn_l and the abbreviation
    r=input_collumns the code should do the following:

    [i=0,...,m_l-1]:
      [j=0,...,r-1]:
        out[j+i \cdot r] = \sum_{k=0}^{n_l-1} (Theta[i,k] * inp[k+j\cdot n_l])

    Arguments:
    ----------
    a_matrix: AMatrix object of size m_l x n_l
    inp: Name of the input vector, treat it like matrix of size n_l x input_rows
    out: Write solut to this vector. Rotation step included.
    input_rows: Current number of rows if you treat inp as matrix.
    """
    # TODO Update documentation
    None


def sumfact_kernel(a_matrix, inp, out, input_collumns, first=False,
                   element=None, name=None, container=None,
                   restriction=None, component=None):
    """Update input_rows and call sumfactorization kernel

    Arguments:
    ----------
    a_matrix: AMatrix object of size m_l x n_l
    inp: Name of the input vector, treat it like matrix of size n_l x input_collumns
    out: Write solut to this vector. Rotation step included.
    input_collumns: Current number of rows if you treat inp as matrix.
    """
    # TODO: Update documentation
    input_collumns /= a_matrix.n
    _sumfact_kernel(a_matrix, inp, out, input_collumns, first, element, name, container, restriction, component)
    print("sumfact kernel, inp: {}, out: {},  input_collumns: {}".format(inp, out, input_collumns))
    input_collumns *= a_matrix.m

    return input_collumns


def sumfact(element, name, container, restriction, component, a_matrices, inp, out):
    # TODO: Think about caching.

    # TODO: Some code suggests that the code works for different n/m
    # per direction (eg. n_1 for x-direction and n_2 for y-direction,
    # of course n!=m is allowed!). This is not tested at all so we
    # assert n=n_i and m=m_i for all i for the moment.
    assert len(a_matrices) > 0
    n = a_matrices[0].n
    m = a_matrices[0].m
    for mat in a_matrices:
        assert n == mat.n
        assert m == mat.m

    # Number of rows of the input of the l-th kernel is equal to:
    #
    # \prod_{i=1}^{l-1} n_i \prod_{i=l+1}^{d} m_i
    #
    # when you do sumfactorization in the order:
    #
    # y = A^{1} ... A^{d} x
    #
    # We just set
    #
    # input_collumns = \prod_{i=1}^{l} n_i
    #
    # and update input_collumns in the sumfact_kernel call!
    input_collumns = 1
    for mat in a_matrices:
        input_collumns *= mat.n

    # Special case for 1D
    if len(a_matrices) == 1:
        input_collumns = sumfact_kernel(a_matrices[0], inp, out, input_collumns,
                                        first=True, element=element, name=name, container=container,
                                        restriction=restriction, component=component)
    else:
        # Create buffer
        buffer_1 = name_buffer(a_matrices)
        if len(a_matrices) > 2:
            buffer_2 = name_buffer(a_matrices)

        if len(a_matrices) % 2 == 0:
            # First kernel
            input_collumns = sumfact_kernel(a_matrices[0], inp, out, input_collumns,
                                            first=True, element=element, name=name, container=container,
                                            restriction=restriction, component=component)

            # In between kernels
            for mat_1, mat_2 in zip(a_matrices[1:-2:2], a_matrices[2:-1:2]):
                input_collumns = sumfact_kernel(mat_1, buffer_1, buffer_2, input_collumns)
                input_collumns = sumfact_kernel(mat_2, buffer_2, buffer_1, input_collumns)

            # Last kernel
            input_collumns = sumfact_kernel(a_matrices[-1], buffer_1, out, input_collumns)
        else:
            # First kernel
            input_collumns = sumfact_kernel(a_matrices[0], inp, out, input_collumns,
                                            first=True, element=element, name=name, container=container,
                                            restriction=restriction, component=component)

            # Second kernel
            input_collumns = sumfact_kernel(a_matrices[1], buffer_1, buffer_2, input_collumns)

            # In between kernels
            for mat_2, mat_1 in zip(a_matrices[2:-2:2], a_matrices[3:-1:2]):
                input_collumns = sumfact_kernel(mat_2, buffer_2, buffer_1, input_collumns)
                input_collumns = sumfact_kernel(mat_1, buffer_1, buffer_2, input_collumns)

            # Last kernel
            input_rowgs = sumfact_kernel(a_matrices[-1], buffer_2, out, input_collumns)
